/*!\file: control_core.cpp
 * \brief: core of the control solution 
 */ 

#include "../toolkits/toolkits.h"
#include "../objects/objects.h"
#include "../shared/shared.h"
#include "../EnumDefinitions/EnumDefinitions.h"
#include "./solutions.h"
#include "../modules/modules.h"
#include "../include/include.h"
#include "../solvers/solvers.h"

void control_core(FemModel* femmodel){

	int     i,n;
	
	/*parameters: */
	int     verbose=0;
	int     control_type;
	int     control_steady;
	int     nsteps;
	double  eps_cm;
	double  tolx;
	double  cm_min;
	double  cm_max;
	int     cm_gradient;

	double* fit=NULL;
	double* optscal=NULL;
	double* maxiter=NULL;
	double* cm_jump=NULL;

		
	/*intermediary: */
	double  search_scalar;
	OptArgs optargs;
	OptPars optpars;

	/*output: */
	double* J=NULL;

	/*Process models*/
	ControlInitialization(femmodel);

	/*Recover parameters used throughout the solution:*/
	femmodel->parameters->FindParam(&nsteps,NStepsEnum);
	femmodel->parameters->FindParam(&control_type,ControlTypeEnum);
	femmodel->parameters->FindParam(&fit,NULL,NULL,FitEnum);
	femmodel->parameters->FindParam(&optscal,NULL,NULL,OptScalEnum);
	femmodel->parameters->FindParam(&maxiter,NULL,NULL,MaxIterEnum);
	femmodel->parameters->FindParam(&cm_jump,NULL,NULL,CmJumpEnum);
	femmodel->parameters->FindParam(&eps_cm,EpsCmEnum);
	femmodel->parameters->FindParam(&tolx,TolXEnum);
	femmodel->parameters->FindParam(&cm_min,CmMinEnum);
	femmodel->parameters->FindParam(&cm_max,CmMaxEnum);
	femmodel->parameters->FindParam(&cm_gradient,CmGradientEnum);
	femmodel->parameters->FindParam(&control_steady,ControlSteadyEnum);

	/*Initialize misfit: */
	J=(double*)xmalloc(nsteps*sizeof(double));
		
	/*Initialize some of the BrentSearch arguments: */
	optargs.femmodel=femmodel;
	optpars.xmin=0; optpars.xmax=1; optpars.tolerance=tolx ;
	
	/*Start looping: */
	for(n=0;n<nsteps;n++){

		_printf_("\n%s%i%s%i\n","   control method step ",n+1,"/",nsteps);
		femmodel->UpdateInputsFromConstant(fit[n],FitEnum);
		
		/*In case we are running a steady state control method, compute new temperature field using new parameter * distribution: */
		if (control_steady) steadystate_core(model);
	
		if(verbose)_printf_("%s\n","      computing gradJ...");
		gradient_core(femmodel);

		/*Return gradient if asked: */
		if (cm_gradient){
			InputToResultx(femodel->elements,femodel->nodes,femodel->vertices,femodel->loads,femodel->materials,femodel->parameters,GradientEnum);
			goto cleanup_and_return;
		}

		_printf_("%s\n","      optimizing along gradient direction");
		optargs.n=n; optpars.maxiter=(int)maxiter[n]; optpars.cm_jump=cm_jump[n];
		BrentSearch(&search_scalar,J+n,&optpars,&objectivefunctionC,&optargs);

		_printf_("%s","      updating parameter using optimized search scalar...");
		AXPYInputx(femmodel->elements,femmodel->nodes,femmodel->vertices,femmodel->loads,femmodel->materials,femmodel->parameters,control_type,search_scalar*optscal[n],ControlParameterEnum);

		_printf_("%s","      constraining the new distribution...");    
		ControlConstrainx(femmodel->elements,femmodel->nodes,femmodel->vertices,femmodel->loads,femmodel->materials,femmodel->parameters,control_type,cm_min,cm_max);
		
		_printf_("%s","      save new parameter...");
		DuplicateInputx(femmodel->elements,femmodel->nodes,femmodel->vertices,femmodel->loads,femmodel->materials,femmodel->parameters,control_type,ControlParameterEnum);
		
		_printf_("%s%i%s%g\n","      value of misfit J after optimization #",n+1,": ",J[n]);
		
		if(controlconvergence)goto cleanup_and_return;

		/*Temporary saving every 5 control steps: */
		if (((n+1)%5)==0){
			_printf_("%s","      saving temporary results...");
			ControlRestart(femmodel);
		}
	}

	_printf_("%s","      preparing final velocity solution");
	if (control_steady) steadystate_core(femmodel);
	else diagnostic_core(femmodel);

	/*some results not computed by steadystate_core or diagnostic_core: */
	InputToResultx(femmodel->elements,femmodel->nodes,femmodel->vertices,femmodel->loads,femmodel->materials,femmodel->parameters,control_type); //the parameter itself!
	femmodel->otherresults->AddObject(new DoubleResult(femmodel->otherresults->Size()+1,0,1,"J",J,nsteps));
	femmodel->otherresults->AddObject(new StringResult(results->Size()+1,ControlTypeEnum,0,1,EnumAsString(control_type)));

	cleanup_and_return:
	
	/*Free ressources: */
	xfree((void**)&control_type);
	xfree((void**)&fit);
	xfree((void**)&optscal);
	xfree((void**)&maxiter);
	xfree((void**)&cm_jump);
	xfree((void**)&J);
}
